;	 <multi media card interface -> mmc functions for atmega8 @ 8MHz>
;    Copyright (C) 2012  Kroetzl Stefan

;    This program is free software: you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation, either version 3 of the License, or
;    (at your option) any later version.

;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.

;    You should have received a copy of the GNU General Public License
;    along with this program.  If not, see <http://www.gnu.org/licenses/>.
;
;	 Contact: http://www.kroetzl.at

;*******************************************************************************************************
;multi_media_card_initialization: 
;parameters: none 
;version: 0.1j
;*******************************************************************************************************

mmc_initialization:

	sbi PORTB, 0x00				; deassert CS

	ldi r17, 0x20

mmc_initialization_sequence:	; at least 74 clock cycles

	ldi r16, 0xFF
	rcall spi_send_byte

	ldi r16, 0x23
	rcall usart_send_byte		; #

	dec r17
	brne mmc_initialization_sequence

mmc_initialization_go_idle_state:				; CMD0

	ldi ZL, low(cmd0*2)							; CMD0
	ldi ZH, high(cmd0*2)

	rcall mmc_initialization_execute

	cpi r16, 0x01	; response must be 0x01
	brne mmc_initialization_go_idle_state

mmc_initialization_crc_on_off:					; CMD59

	ldi ZL, low(cmd59*2)						; CMD59
	ldi ZH, high(cmd59*2)

	rcall mmc_initialization_execute

	cpi r16, 0x01	; still in idle state
	brne mmc_initialization_crc_on_off

mmc_initialization_read_ocr:					; CMD58

	rcall mmc_read_ocr							; CMD58

	cpi r16, 0x01
	brne mmc_initialization_read_ocr			

mmc_initialization_send_operation_conditions:	; CMD1

	ldi ZL, low(cmd1*2)							; CMD1
	ldi ZH, high(cmd1*2)

	rcall mmc_initialization_execute

	cpi r16,0x00
	brne mmc_initialization_send_operation_conditions

mmc_initialization_send_csd:					; CMD9

	ldi ZL, low(cmd9*2)
	ldi ZH, high(cmd9*2)

	rcall mmc_initialization_execute

	cpi r16, 0x00
	brne mmc_initialization_send_csd

	rcall mmc_read_register

mmc_initialization_send_cid:					; CMD10

	ldi ZL, low(cmd10*2)
	ldi ZH, high(cmd10*2)

	rcall mmc_initialization_execute
	
	cpi r16, 0x00
	brne mmc_initialization_send_cid

	rcall mmc_read_register

mmc_initialization_ready:

	ret

mmc_initialization_execute:		; make sure the Z pointer is set properly

	rcall usart_send_string		; transmit the ascii part
	rcall mmc_program_memory_to_sram	; load the binary part into the sram
	rcall mmc_send_command		; send the command
	rcall usart_int_to_ascii	; show the response

	ret

;*******************************************************************************************************
;mmc_program_memory_to_sram:
;parameters: Z
;version: 0.1a
;*******************************************************************************************************

mmc_program_memory_to_sram:

	ldi YL, low(mmc)	; inizialize Y pointer with mmc base adress
	ldi YH, high(mmc)

	ldi r17, 0x06

mmc_program_memory_to_sram_shift:

	lpm r16, Z+
	st Y+, r16

	dec r17
	brne mmc_program_memory_to_sram_shift

	ret	

;*******************************************************************************************************
;mmc_send_command:
;parameters: none
;version: 0.1a
;*******************************************************************************************************

mmc_send_command:

	ldi ZL, low(mmc)
	ldi ZH, high(mmc)

	cbi PORTB, 0x00		; assert CS

	ldi r17, 0x06
		
mmc_send_command_loop:	; fetch from sram and transmit

	ld r16, Z+
	rcall spi_send_byte
	
	dec r17
	brne mmc_send_command_loop

mmc_send_command_response:

	cpi r16, 0xFF
	brne mmc_send_command_return

	ldi r16,0xFF
	rcall spi_send_byte

	rjmp mmc_send_command_response

mmc_send_command_return:

	push r16

	ldi r16, 0xFF
	rcall spi_send_byte

	mov r17, r16		; save token -> r2 response or start token
; make sure that you don't alter the content of r17 before rcall mmc_poll_bytes
	pop r16				
							; 
	sbi PORTB, 0x00		; deassert CS

	ret

;*******************************************************************************************************
;mmc_read_block:
;parameters:
;version: 0.1a
;*******************************************************************************************************

mmc_read_block:

	ldi ZL, low(mmc)							
	ldi ZH, high(mmc)

	ldi r16, 0x51	; CMD17
	st Z+, r16

	ldi r16, 0x00
	st Z+, r16
	st Z+, r16
	st Z+, r16
	st Z+, r16
	st Z+, r16

	rcall crc7

	rcall mmc_send_command

	cpi r16, 0x00										
	brne mmc_read_block

	ldi YL, low(stack)
	ldi YH, high(stack)

	rcall mmc_poll_bytes

	ret

;*******************************************************************************************************
;mmc_read_register:
;parameters:
;version: 0.1a
;*******************************************************************************************************

mmc_read_register:

	ldi YL, low(cidcsd)
	ldi YH, high(cidcsd)

	rcall mmc_poll_bytes

	ret

;*******************************************************************************************************
;mmc_poll_bytes: 
;parameters: Y, r17
;version: 0.1h
;*******************************************************************************************************

mmc_poll_bytes:			; some cards transmit the start token immediately after transmitting the r1
						; response. make sure that you don't alter the content of r17 before 
	ldi ZL, low(mmc)	; calling mmc_poll_bytes
	ldi ZH, high(mmc)

	mov r16, r17		; load the second response of mmc_send_command

	clr r18				; crc16 16bit register initialization
	clr r19

	cbi PORTB, 0x00		; assert CS

mmc_poll_bytes_start_token:

	cpi r16, 0xFE
	breq mmc_poll_bytes_loop	; check for start byte

	ldi r16, 0xFF		; dummy token
	rcall spi_send_byte	; read byte

	rjmp mmc_poll_bytes_start_token

mmc_poll_bytes_loop:

	ldi r16, 0xFF 		; dummy token
	rcall spi_send_byte	; read Byte

	rcall usart_int_to_ascii

	st Z+, r16
	rcall crc16

	cp ZL, YL			; transmision is complete?
	cpc ZH, YH
	brlo mmc_poll_bytes_loop

	ldi r16, 0x00		; the input data is augmented by 16 zero bits
	rcall crc16		

	ldi r16, 0x00
	rcall crc16			; crc16 calculation complete

	ldi r16, 0xFF		; fetch crc16 bits ...
	rcall spi_send_byte

	mov r23, r16

	ldi r16, 0xFF
	rcall spi_send_byte	

	mov r22, r16		; ... done

	cp r18, r22			; crc16 check
	cpc r19, r23
	brne mmc_poll_bytes_fail

	ldi r16, 0xFF
	rcall spi_send_byte	; additional clock cycles

mmc_poll_bytes_return:

	sbi PORTB, 0x00		; deassert CS

	ret

mmc_poll_bytes_fail:

	ldi ZL, low(crc*2)
	ldi ZH, high(crc*2)

	rcall usart_send_string

	rjmp mmc_poll_bytes_fail

;*******************************************************************************************************
;mmc_read_ocr: since we get a r3 response we need a own function for this
;parameters:
;version: 0.1a
;*******************************************************************************************************

mmc_read_ocr:

	ldi ZL, low(cmd58*2)	; CMD58
	ldi ZH, high(cmd58*2)

	rcall usart_send_string

	cbi PORTB, 0x00		; assert CS

	ldi r17, 0x06
		
mmc_read_ocr_loop:

	lpm r16, Z+
	rcall spi_send_byte
	
	dec r17
	brne mmc_read_ocr_loop

mmc_read_ocr_response:

	cpi r16, 0xFF
	brne mmc_read_ocr_return

	ldi r16,0xFF
	rcall spi_send_byte

	rjmp mmc_read_ocr_response

mmc_read_ocr_return:

	ldi ZL, low(mmc)
	ldi ZH, high(mmc)

	st Z+, r16
	rcall usart_int_to_ascii

	ldi r17, 0x04		; r3 response

	push r16

mmc_read_ocr_return_loop:

	ldi r16, 0xFF
	rcall spi_send_byte

	push r17

	st Z+, r16
	rcall usart_int_to_ascii

	pop r17

	dec r17
	brne mmc_read_ocr_return_loop

	pop r16

	sbi PORTB, 0x00		; deassert CS

	ret	

