;	 <multi media card interface -> usart for atmega8 @ 8MHz>
;    Copyright (C) 2012  Kroetzl Stefan

;    This program is free software: you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation, either version 3 of the License.

;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.

;    You should have received a copy of the GNU General Public License
;    along with this program.  If not, see <http://www.gnu.org/licenses/>.
;
;	 Contact: http://www.kroetzl.at

;*******************************************************************************************************
;usart_initialization:	38400 8 N 1 flow control: off	@8Mhz  
;parameters: none 
;version: 0.1h
;*******************************************************************************************************

usart_initialization:

	ldi r16, 0x0C		; 
	out UBRRL, r16		
	ldi r16, 0x00
	out UBRRH, r16

	sbi UCSRB, TXEN
	sbi UCSRB, RXEN

	ldi ZL, low(licence*2)
	ldi ZH, high(licence*2)

	rcall usart_send_string

usart_initialization_loop:

	sbis UCSRA, RXC
	rjmp usart_initialization_loop

	in r16, UDR ; not used

	rcall usart_carriage_return

	ret

;*******************************************************************************************************
;usart_send_byte:
;parameters: none 
;version: 0.1a
;*******************************************************************************************************

usart_send_byte:	
							
	sbis UCSRA, UDRE
	rjmp usart_send_byte

	out UDR, r16

	ret

;*******************************************************************************************************
;usart_receive_byte:
;parameters: none 
;version: 0.1h
;*******************************************************************************************************

usart_receive_byte:

	sbis UCSRA, RXC
	rjmp usart_receive_byte
   
	in r16, UDR

	rcall usart_send_byte

usart_receive_byte_return:

	ret

;*******************************************************************************************************
;usart_receive_shell:
;parameters: none 
;version: 0.1a
;*******************************************************************************************************
 
usart_shell:	; this is the command shell.

	rcall usart_receive_string	; first of all we need a string

	ldi ZL, low(instruction*2)
	ldi ZH, high(instruction*2)

	rcall usart_receive_process_string	; and now we can compare that string with the available commands
										; and see if we get any matches.
	ldi ZL, low(poll*2)
	ldi ZH, high(poll*2)

	rcall usart_receive_process_string

	ldi ZL, low(write*2)
	ldi ZH, high(write*2)

	rcall usart_receive_process_string

	ret

;*******************************************************************************************************
;usart_receive_string:
;parameters: none 
;version: 0.1a
;*******************************************************************************************************

usart_receive_string:

	ldi YL, low(mmc)
	ldi YH, high(mmc)

	ldi XL, low(cidcsd)		; max instruction length 16 bytes!
	ldi XH, high(cidcsd)

	push r16

usart_receive_string_loop:

	rcall usart_receive_byte

	st Y+, r16
	
	cpi r16, 0x0D
	breq usart_receive_string_return	; done

	cp YL, XL
	cpc YH, XH
	brlo usart_receive_string_loop	; instruction exceeded length ?	

usart_receive_string_return:

	pop r16

	ret

;*******************************************************************************************************
;usart_receive_process_string:
;parameters: none 
;version: 0.1a
;*******************************************************************************************************

usart_receive_process_string:

	push r16
	push r17

	ldi YL, low(mmc)
	ldi YH, high(mmc)

usart_receive_process_loop:	; we just compare the bytes fetched from the usart and those stored in the 
							; program memory. if they are equal everthing is fine.
	ld r16, Y+
	lpm r17, Z+

	cpi r16, 0x0D			; a carriage return marks the end of the command
	breq usart_receive_process_ijmp

	cpi r17, 0x00			; actually this is bad because there was no carriage return -> abort
	breq usart_receive_process_return

	cp r16, r17
	breq usart_receive_process_loop

usart_receive_process_return:	; not equal -> maby we are more lucky with another string

	pop r17		
	pop r16

 	ret 

usart_receive_process_ijmp:

	cpi r17, 0x00			; no we check the byte from the program memory
	brne usart_receive_process_return	; if it is 0x00 the strings are identical

	lpm r16, Z+	; behind the string we can find the adress of the function
	lpm r17, Z

	mov ZL, r16
	mov ZH, r17

	pop r17
	pop r16

	ijmp		; jump into the function

	ret	; just in case something goes seriously wrong

;*******************************************************************************************************
;usart_receive_set_byte:
;parameters: Z 
;version: 0.1a
;*******************************************************************************************************

usart_receive_set_byte:

	push r16
	rcall usart_int_to_ascii

	ldi r16,0x0D
	rcall usart_send_byte	; reset the cursor

usart_receive_set_byte_loop:

	sbis UCSRA, RXC
	rjmp usart_receive_set_byte_loop

	in r16, UDR

	cpi r16, 0x0D
	breq usart_receive_set_byte_return		; the user hit return -> byte is ready 

	cpi r16, 0x2B
	breq usart_receive_set_byte_increase	; got a +

	cpi r16, 0x2D
	breq usart_receive_set_byte_decrease	; got a -
	
	pop r16
	
	rjmp usart_receive_set_byte				; got something else

usart_receive_set_byte_return:	

	pop r16
	st Z+, r16				; write byte to sram

	ret

usart_receive_set_byte_increase:

	pop r16
	inc r16

	rjmp usart_receive_set_byte

usart_receive_set_byte_decrease:

	pop r16
	dec r16

	rjmp usart_receive_set_byte

;*******************************************************************************************************
;usart_receive_instruction:
;parameters: none 
;version: 0.1a
;*******************************************************************************************************

usart_receive_instruction:

	ldi r16, 0x0A
	rcall usart_send_byte

	ldi ZL, low(mmc)
	ldi ZH, high(mmc)

	ldi r17, 0x06	; mmc command is 5 bytes long + 1 crc7 byte

usart_receive_instruction_loop:

	ldi r16, 0x00
	rcall usart_receive_set_byte

	rcall usart_carriage_return

	dec r17
	brne usart_receive_instruction_loop
	
	rcall usart_carriage_return
	rcall crc7
	rcall mmc_send_command
	rcall usart_int_to_ascii
	rcall usart_carriage_return

	ret

;*******************************************************************************************************
;usart_receive_poll:
;parameters: none 
;version: 0.1a
;*******************************************************************************************************

usart_receive_poll:

	ldi r16, 0x0A
	rcall usart_send_byte

	ldi ZL, low(mmc)
	ldi ZH, high(mmc)

	ldi r16, 0x60		; base adress is 0x60 don't try anything lower than that.
	rcall usart_receive_set_byte

	ld YL, -Z			; make sure you set the correct length otherwise you get a crc error

	rcall usart_carriage_return

	ldi r16, 0x00
	rcall usart_receive_set_byte

	ld YH, -Z

	rcall usart_carriage_return
	rcall usart_carriage_return

	ldi XL, low(cidcsd)
	ldi XH, high(cidcsd)	; minimum 16 bytes !

	cp YL, XL
	cpc YH, XH
	brlo usart_receive_poll_error

	ldi XL, low(stack)
	ldi XH, high(stack)

	cp XL, YL
	cpc XH, YH
	brlo usart_receive_poll_error	; maximum 512bytes ! 

	rcall mmc_poll_bytes

	rcall usart_carriage_return

	ldi ZL, low(mmc)
	ldi ZH, high(mmc)

usart_receive_poll_binary:

	ld r16, Z+
	rcall usart_send_byte

	cp ZL, YL
	cpc ZH, YH
	brne usart_receive_poll_binary

usart_receive_poll_return:

	rcall usart_carriage_return

	ret

usart_receive_poll_error:

	rcall usart_carriage_return

	ldi ZL, low(address*2)
	ldi ZH, high(address*2)

	rcall usart_send_string

	rjmp usart_receive_poll_return

;*******************************************************************************************************
;usart_receive_write:
;parameters: none 
;version: 0.1a
;*******************************************************************************************************

usart_receive_write: ; not implemented

	ret

;*******************************************************************************************************
;usart_send_string:
;parameters: none 
;version: 0.1a
;*******************************************************************************************************

usart_send_string:

    lpm		r16, Z+
    and     r16,r16                   
    breq    usart_string_return

	rcall usart_send_byte	      

    rjmp    usart_send_string           

usart_string_return:

    ret

;*******************************************************************************************************
;usart_int_to_ascii: www.mikrocontroller.net -> AVR-Tutorial
;parameters: r16
;version: 0.1a
;*******************************************************************************************************

usart_int_to_ascii:

	push r16
	push r17

	mov r17, r16

	ldi r16, '0'-1       
usart_int_to_ascii_hundret:

    inc r16
    subi r17, 100	    ; -100
 
    brcc usart_int_to_ascii_hundret
                                    
    subi r17, -100	    ; +100

	rcall usart_send_byte
 
    ldi r16, '0'-1      ;
usart_int_to_ascii_ten:

    inc r16
    subi r17, 10      	; -10
    
    brcc usart_int_to_ascii_ten
                                    
    subi r17, -10		; +10

	rcall usart_send_byte
 
    subi r17, -'0'
	mov r16, r17

	rcall usart_send_byte

	ldi r16, 0x20
	rcall usart_send_byte

	pop r17
	pop r16

	ret

;*******************************************************************************************************
;usart_carriage_return:
;parameters: none 
;version: 0.1a
;*******************************************************************************************************

usart_carriage_return:

	ldi r16, 0x0D
	rcall usart_send_byte
	ldi r16, 0x0A
	rcall usart_send_byte

	ret
